"""Nf-tower related constants."""

from typing import Any

NALLO_GENERAL_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "median_coverage": {"norm": "gt", "threshold": 20},
}

NALLO_PARENT_PEDDY_METRIC_CONDITION: dict[str, dict[str, Any]] = {
    "parent_error_ped_check": {"norm": "eq", "threshold": "False"},
}

NALLO_RAW_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "somalier_sex": {"norm": "eq", "threshold": None},
}

RAREDISEASE_PREDICTED_SEX_METRIC = "predicted_sex_sex_check"

RAREDISEASE_METRIC_CONDITIONS_WES: dict[str, dict[str, Any]] = {
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.20},
    "PCT_PF_UQ_READS_ALIGNED": {"norm": "gt", "threshold": 0.95},
    "PCT_TARGET_BASES_10X": {"norm": "gt", "threshold": 0.95},
    "AT_DROPOUT": {"norm": "lt", "threshold": 10},
    "GC_DROPOUT": {"norm": "lt", "threshold": 10},
    RAREDISEASE_PREDICTED_SEX_METRIC: {"norm": "eq", "threshold": None},
    "gender": {"norm": "eq", "threshold": None},
    "FREEMIX": {"norm": "lt", "threshold": 0.02},
}

RAREDISEASE_METRIC_CONDITIONS_WGS: dict[str, dict[str, Any]] = {
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.20},
    "PCT_PF_UQ_READS_ALIGNED": {"norm": "gt", "threshold": 0.95},
    "MEDIAN_TARGET_COVERAGE": {"norm": "gt", "threshold": 25},
    "PCT_TARGET_BASES_10X": {"norm": "gt", "threshold": 0.95},
    "AT_DROPOUT": {"norm": "lt", "threshold": 5},
    "GC_DROPOUT": {"norm": "lt", "threshold": 5},
    RAREDISEASE_PREDICTED_SEX_METRIC: {"norm": "eq", "threshold": None},
    "gender": {"norm": "eq", "threshold": None},
    "FREEMIX": {"norm": "lt", "threshold": 0.02},
}

RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION: dict[str, dict[str, Any]] = {
    "parent_error_ped_check": {"norm": "eq", "threshold": "False"},
}

RNAFUSION_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.7},
}

RNAFUSION_METRIC_CONDITIONS_DEPLETION: dict[str, dict[str, Any]] = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 0},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.7},
}

RNAFUSION_METRIC_CONDITIONS_FOR_APPTAG: dict[str, dict[str, dict[str, Any]]] = {
    "RNAWDPR100": RNAFUSION_METRIC_CONDITIONS_DEPLETION
}

TOMTE_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "pct_duplication": {"norm": "lt", "threshold": 70},
}

NALLO_COVERAGE_FILE_TAGS: list[str] = ["d4"]
NALLO_COVERAGE_INTERVAL_TYPE: str = "genes"
NALLO_COVERAGE_THRESHOLD: int = 10

RAREDISEASE_COVERAGE_FILE_TAGS: list[str] = ["coverage", "d4"]



from cg.services.orders.validation.order_types.balsamic.models.order import BalsamicOrder
from cg.services.orders.validation.order_types.balsamic_umi.constants import BalsamicUmiDeliveryType


class BalsamicUmiOrder(BalsamicOrder):
    delivery_type: BalsamicUmiDeliveryType



import coloredlogs
import requests
from flask import Flask, redirect, session, url_for
from flask_admin.base import AdminIndexView
from flask_dance.consumer import oauth_authorized
from flask_dance.contrib.google import google, make_google_blueprint
from sqlalchemy.orm import scoped_session

from cg.server import admin, ext, invoices
from cg.server.app_config import app_config
from cg.server.endpoints.analyses import ANALYSES_BLUEPRINT
from cg.server.endpoints.applications import APPLICATIONS_BLUEPRINT
from cg.server.endpoints.cases import CASES_BLUEPRINT
from cg.server.endpoints.index_sequences import INDEX_SEQUENCES_BLUEPRINT
from cg.server.endpoints.orders import ORDERS_BLUEPRINT
from cg.server.endpoints.pools import POOLS_BLUEPRINT
from cg.server.endpoints.samples import SAMPLES_BLUEPRINT
from cg.server.endpoints.sequencing_metrics.illumina_sequencing_metrics import FLOW_CELLS_BLUEPRINT
from cg.server.endpoints.sequencing_metrics.pacbio_sequencing_metrics import (
    PACBIO_SAMPLE_SEQUENCING_METRICS_BLUEPRINT,
)
from cg.server.endpoints.sequencing_run.pacbio_sequencing_run import (
    PACBIO_SEQUENCING_RUNS_BLUEPRINT,
)
from cg.server.endpoints.sequencing_run.pacbio_smrt_cell_metrics import (
    PACBIO_SMRT_CELL_METRICS_BLUEPRINT,
)
from cg.server.endpoints.users import USERS_BLUEPRINT
from cg.store.database import get_scoped_session_registry
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    ApplicationVersion,
    Bed,
    BedVersion,
    Case,
    CaseSample,
    Collaboration,
    Customer,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Invoice,
    Order,
    Organism,
    PacbioSampleSequencingMetrics,
    PacbioSMRTCellMetrics,
    Panel,
    Pool,
    Sample,
    User,
)


def create_app():
    """Generate a flask application."""
    app = Flask(__name__, template_folder="templates")
    _load_config(app)
    _configure_extensions(app)
    _register_blueprints(app)
    _register_teardowns(app)

    return app


def _load_config(app: Flask):
    app.config.update(app_config.model_dump())
    app.secret_key = app_config.cg_secret_key


def _configure_extensions(app: Flask):
    _initialize_logging(app)
    certs_resp = requests.get("https://www.googleapis.com/oauth2/v1/certs")
    app.config["GOOGLE_OAUTH_CERTS"] = certs_resp.json()

    ext.cors.init_app(app)
    ext.csrf.init_app(app)
    ext.db.init_app(app)
    ext.lims.init_app(app)
    ext.analysis_client.init_app(app)
    ext.admin.init_app(app, index_view=AdminIndexView(endpoint="admin"))
    app.json_provider_class = ext.CustomJSONEncoder


def _initialize_logging(app):
    coloredlogs.install(level="DEBUG" if app.debug else "INFO")


def _register_blueprints(app: Flask):
    oauth_bp = make_google_blueprint(
        client_id=app_config.google_oauth_client_id,
        client_secret=app_config.google_oauth_client_secret,
        scope=["openid", "https://www.googleapis.com/auth/userinfo.email"],
    )

    @oauth_authorized.connect_via(oauth_bp)
    def logged_in(blueprint, token):
        """Called when the user logs in via Google OAuth."""
        resp = google.get("/oauth2/v1/userinfo?alt=json")
        assert resp.ok, resp.text
        user_data = resp.json()
        session["user_email"] = user_data["email"]

    app.register_blueprint(invoices.BLUEPRINT, url_prefix="/invoices")
    app.register_blueprint(oauth_bp, url_prefix="/login")
    app.register_blueprint(APPLICATIONS_BLUEPRINT)
    app.register_blueprint(CASES_BLUEPRINT)
    app.register_blueprint(ORDERS_BLUEPRINT)
    app.register_blueprint(SAMPLES_BLUEPRINT)
    app.register_blueprint(POOLS_BLUEPRINT)
    app.register_blueprint(FLOW_CELLS_BLUEPRINT)
    app.register_blueprint(ANALYSES_BLUEPRINT)
    app.register_blueprint(USERS_BLUEPRINT)
    app.register_blueprint(PACBIO_SAMPLE_SEQUENCING_METRICS_BLUEPRINT)
    app.register_blueprint(PACBIO_SEQUENCING_RUNS_BLUEPRINT)
    app.register_blueprint(PACBIO_SMRT_CELL_METRICS_BLUEPRINT)
    app.register_blueprint(INDEX_SEQUENCES_BLUEPRINT)
    _register_admin_views()

    ext.csrf.exempt(SAMPLES_BLUEPRINT)
    ext.csrf.exempt(CASES_BLUEPRINT)
    ext.csrf.exempt(APPLICATIONS_BLUEPRINT)
    ext.csrf.exempt(ORDERS_BLUEPRINT)
    ext.csrf.exempt(POOLS_BLUEPRINT)
    ext.csrf.exempt(FLOW_CELLS_BLUEPRINT)
    ext.csrf.exempt(ANALYSES_BLUEPRINT)
    ext.csrf.exempt(USERS_BLUEPRINT)
    ext.csrf.exempt(PACBIO_SEQUENCING_RUNS_BLUEPRINT)

    @app.route("/")
    def index():
        return redirect(url_for("admin.index"))

    @app.route("/logout")
    def logout():
        """Log out the user."""
        session["user_email"] = None
        return redirect(url_for("index"))


def _register_admin_views():
    # Base data views
    ext.admin.add_view(admin.ApplicationView(Application, ext.db.session))
    ext.admin.add_view(admin.ApplicationVersionView(ApplicationVersion, ext.db.session))
    ext.admin.add_view(admin.ApplicationLimitationsView(ApplicationLimitations, ext.db.session))
    ext.admin.add_view(admin.BedView(Bed, ext.db.session))
    ext.admin.add_view(admin.BedVersionView(BedVersion, ext.db.session))
    ext.admin.add_view(admin.CustomerView(Customer, ext.db.session))
    ext.admin.add_view(admin.CollaborationView(Collaboration, ext.db.session))
    ext.admin.add_view(admin.OrganismView(Organism, ext.db.session))
    ext.admin.add_view(admin.OrderView(Order, ext.db.session))
    ext.admin.add_view(admin.PanelView(Panel, ext.db.session))
    ext.admin.add_view(admin.UserView(User, ext.db.session))

    # Business data views
    ext.admin.add_view(admin.CaseView(Case, ext.db.session))
    ext.admin.add_view(admin.CaseSampleView(CaseSample, ext.db.session))
    ext.admin.add_view(admin.SampleView(Sample, ext.db.session))
    ext.admin.add_view(admin.PoolView(Pool, ext.db.session))
    ext.admin.add_view(admin.AnalysisView(Analysis, ext.db.session))
    ext.admin.add_view(admin.InvoiceView(Invoice, ext.db.session))
    ext.admin.add_view(
        admin.IlluminaFlowCellView(IlluminaSequencingRun, ext.db.session, name="Illumina Flow Cell")
    )
    ext.admin.add_view(
        admin.IlluminaSampleSequencingMetricsView(IlluminaSampleSequencingMetrics, ext.db.session)
    )
    ext.admin.add_view(
        admin.PacbioSmrtCellMetricsView(
            PacbioSMRTCellMetrics, ext.db.session, "Pacbio SMRT Cell Metrics"
        )
    )
    ext.admin.add_view(
        admin.PacbioSampleRunMetricsView(PacbioSampleSequencingMetrics, ext.db.session)
    )


def _register_teardowns(app: Flask):
    """Register teardown functions."""

    @app.teardown_appcontext
    def remove_database_session(exception=None):
        """
        Remove the database session to ensure database resources are
        released when a request has been processed.
        """
        scoped_session_registry: scoped_session | None = get_scoped_session_registry()
        if scoped_session_registry:
            scoped_session_registry.remove()



"""Module for Nallo Analysis API."""

import copy
import logging
from itertools import permutations
from pathlib import Path
from typing import Any

from click import File

from cg.clients.chanjo2.models import (
    CoverageMetrics,
    CoveragePostRequest,
    CoveragePostResponse,
    CoverageSample,
)
from cg.constants import Workflow
from cg.constants.constants import GenomeVersion
from cg.constants.nf_analysis import (
    NALLO_COVERAGE_FILE_TAGS,
    NALLO_COVERAGE_INTERVAL_TYPE,
    NALLO_COVERAGE_THRESHOLD,
    NALLO_GENERAL_METRIC_CONDITIONS,
    NALLO_PARENT_PEDDY_METRIC_CONDITION,
    NALLO_RAW_METRIC_CONDITIONS,
)
from cg.constants.scout import NALLO_CASE_TAGS
from cg.constants.subject import PlinkSex
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.nallo.nallo import NalloQCMetrics
from cg.resources import NALLO_BUNDLE_FILENAMES_PATH
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


class NalloAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Nallo processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.NALLO,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.nallo.root
        self.workflow_bin_path: str = config.nallo.workflow_bin_path
        self.profile: str = config.nallo.profile
        self.conda_env: str = config.nallo.conda_env
        self.conda_binary: str = config.nallo.conda_binary
        self.platform: str = config.nallo.platform
        self.params: str = config.nallo.params
        self.workflow_config_path: str = config.nallo.config
        self.resources: str = config.nallo.resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.nallo.tower_workflow
        self.account: str = config.nallo.slurm.account
        self.email: str = config.nallo.slurm.mail_user
        self.revision: str = config.nallo.revision

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a Nallo case. Currently fixed for hg38."""
        return GenomeVersion.HG38

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Nallo bundle filenames path."""
        return NALLO_BUNDLE_FILENAMES_PATH

    def get_qc_conditions_for_workflow(self, sample_id: str) -> dict:
        """Return Nallo workflow metric conditions for a sample."""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        if "-" not in sample_id:
            metric_conditions: dict[str, dict[str, Any]] = NALLO_GENERAL_METRIC_CONDITIONS.copy()
            metric_conditions.update(NALLO_RAW_METRIC_CONDITIONS)
            self.set_somalier_sex_for_sample(sample=sample, metric_conditions=metric_conditions)
        else:
            metric_conditions = NALLO_PARENT_PEDDY_METRIC_CONDITION.copy()
        return metric_conditions

    def _get_nallo_sample_pair_patterns(self, case_id: str) -> list[str]:
        """Return sample-pair patterns for searching in MultiQC."""
        sample_ids: list[str] = list(self.status_db.get_sample_ids_by_case_id(case_id=case_id))
        pairwise_patterns: list[str] = [
            f"{sample1}-{sample2}" for sample1, sample2 in permutations(sample_ids, 2)
        ]
        return pairwise_patterns

    def get_nallo_parent_error_ped_check_metric(
        self, pair_sample_ids: str, multiqc_raw_data: dict[dict]
    ) -> MetricsBase | None:
        """Return the parsed metrics for pedigree error given a concatenated pair of sample ids."""
        metric_name: str = "parent_error_ped_check"
        peddy_metrics: dict[str, dict] = multiqc_raw_data["multiqc_peddy"]
        if sample_pair_metrics := peddy_metrics.get(pair_sample_ids, None):
            return self.get_multiqc_metric(
                metric_name=metric_name,
                metric_value=sample_pair_metrics[metric_name],
                sample_id=pair_sample_ids,
            )

    def get_nallo_raw_metric(self, sample_id: str, multiqc_raw_data: dict) -> MetricsBase | None:
        metric_name = "sex"
        raw_metrics_section: dict[str, dict] = multiqc_raw_data.get("multiqc_somalier", {})
        sample_metrics = raw_metrics_section.get(sample_id)
        if sample_metrics and metric_name in sample_metrics:
            return self.get_multiqc_metric(
                metric_name="somalier_sex",
                metric_value=sample_metrics[metric_name],
                sample_id=sample_id,
            )

    def get_nallo_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the Nallo metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []

        for pattern in self.get_multiqc_search_patterns(case_id):
            metrics_for_pattern: list[MetricsBase] = self.get_multiqc_metrics_for_sample(
                search_pattern=pattern.pattern,
                multiqc_json=multiqc_json,
                sample_id=pattern.sample_id,
                exact_match=self.is_multiqc_pattern_search_exact,
            )
            metrics.extend(metrics_for_pattern)
        for sample_id in self.status_db.get_sample_ids_by_case_id(case_id):
            if raw_metric := self.get_nallo_raw_metric(
                sample_id=sample_id,
                multiqc_raw_data=multiqc_json.report_saved_raw_data,
            ):
                metrics.append(raw_metric)
        for sample_pair in self._get_nallo_sample_pair_patterns(case_id):
            if parent_error_metric := self.get_nallo_parent_error_ped_check_metric(
                pair_sample_ids=sample_pair, multiqc_raw_data=multiqc_json.report_saved_raw_data
            ):
                metrics.append(parent_error_metric)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    def _get_list_of_metric_dicts(self, multiqc_json: MultiqcDataJson) -> list[dict[str, Any]]:
        metric_dicts: list[dict[str, Any]] = super()._get_list_of_metric_dicts(multiqc_json)

        list_copy: list[dict[str, Any]] = copy.deepcopy(metric_dicts)
        list_copy.append(self._get_somalier_dict(multiqc_json))

        return list_copy

    def _get_somalier_dict(self, multiqc_json: MultiqcDataJson) -> dict[str, Any]:
        somalier_raw = copy.deepcopy(multiqc_json.report_saved_raw_data["multiqc_somalier"])

        for sample_id, metrics in somalier_raw.items():
            somalier_raw[sample_id] = {f"somalier_{k}": v for k, v in metrics.items()}

        return somalier_raw

    @staticmethod
    def set_somalier_sex_for_sample(sample: Sample, metric_conditions: dict) -> None:
        if "somalier_sex" in metric_conditions:
            metric_conditions["somalier_sex"]["threshold"] = float(
                {
                    "male": PlinkSex.MALE,
                    "female": PlinkSex.FEMALE,
                    "unknown": PlinkSex.UNKNOWN,
                }[sample.sex]
            )

    def get_sample_coverage_file_path(self, bundle_name: str, sample_id: str) -> str | None:
        """Return the Nallo d4 coverage file path."""
        nallo_coverage_file_tags: list[str] = NALLO_COVERAGE_FILE_TAGS + [sample_id]
        nallo_coverage_file: File | None = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=bundle_name, tags=nallo_coverage_file_tags
        )
        if nallo_coverage_file:
            return nallo_coverage_file.full_path
        LOG.warning(f"No Nallo coverage file found with the tags: {nallo_coverage_file_tags}")
        return None

    def get_sample_coverage(
        self, case_id: str, sample_id: str, gene_ids: list[int]
    ) -> CoverageMetrics | None:
        """Return sample coverage metrics from Chanjo2."""
        nallo_genome_version: GenomeVersion = self.get_genome_build(case_id)
        nallo_coverage_file_path: str | None = self.get_sample_coverage_file_path(
            bundle_name=case_id, sample_id=sample_id
        )
        try:
            post_request = CoveragePostRequest(
                build=self.translate_genome_reference(nallo_genome_version),
                coverage_threshold=NALLO_COVERAGE_THRESHOLD,
                hgnc_gene_ids=gene_ids,
                interval_type=NALLO_COVERAGE_INTERVAL_TYPE,
                samples=[
                    CoverageSample(coverage_file_path=nallo_coverage_file_path, name=sample_id)
                ],
            )
            post_response: CoveragePostResponse = self.chanjo2_api.get_coverage(post_request)
            return post_response.get_sample_coverage_metrics(sample_id)
        except Exception as error:
            LOG.error(f"Error getting coverage for sample '{sample_id}', error: {error}")
            return None

    def get_scout_upload_case_tags(self) -> dict:
        """Return Nallo Scout upload case tags."""
        return NALLO_CASE_TAGS

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        qc_metrics_model = NalloQCMetrics
        return super().parse_analysis(
            qc_metrics_raw=qc_metrics_raw, qc_metrics_model=qc_metrics_model, **kwargs
        )

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of a Nallo metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_nallo_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}
